//  ----------------------------------------------------------------------
// | Open Digital Libraries Component Explorer - version 1.0              |
// | Hussein Suleman                                                      |
// | September 2002                                                       |
//  ----------------------------------------------------------------------
// |  Virginia Polytechnic Institute and State University                 |
// |  Department of Computer Science                                      |
// |  Digital Library Research Laboratory                                 |
//  ----------------------------------------------------------------------

// ======================================================================
//  Module     : addarchive
//  Purpose    : add a new archive to list of predefined archives
// ======================================================================

#include <fstream.h>
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "blocks.h"
#include "cgi.h"
#include "config.h"
#include "language.h"
#include "getproto.h"

// ----------------------------------------------------------------------
//  Function   : ScreenOne
// ----------------------------------------------------------------------
//  Purpose    : generate screen to capture archive url
//  Parameters : 1. language name
//               2. bgcolor
//               3. headercolor
//               4. blockcolor
//  Result     : (none)
// ----------------------------------------------------------------------

void ScreenOne ( char *language, char *bgcolor, char *headercolor, 
                 char *blockcolor )
{
   cout << "<h2>" << Translate ("Test and Add an archive or component")
        << "</h2><hr>" 
        << Translate ("Please note that these tests are neither definitive nor exhaustive")
        << ". " << Translate ("They serve not as a proof of correctness of implementation, but rather indicate possible errors if they exist")
        << ". " << Translate ("If you have any questions please contact the author at the address below")
        << ".<hr>";

   cout << Translate ("If your component supports XOAI-PMH v1.0 (ODL protocols)") << " ... "
        << Translate ("Enter the base URL of the component") << " :<p>"
        << "<form method=\"get\" action=\"" << addarchive << "\">"
        << "<input type=text name=archiveurl size=50 maxlength=200>"
        << "<input type=hidden name=language value=\"" << language << "\">"
        << "<input type=hidden name=bgcolor value=\"" << bgcolor << "\">"
        << "<input type=hidden name=headercolor value=\"" << headercolor << "\">"
        << "<input type=hidden name=blockcolor value=\"" << blockcolor << "\">"
        << "<input type=hidden name=gotarchive value=\"gotarchive\">"
        << "<input type=hidden name=version value=\"xoai10\">"
        << "<input type=submit name=submit value=\""
        << Translate ("Test the component") << "\">"
        << "</form><hr>";

   cout << Translate ("If your archive supports OAI-PMH v2.0") << " ... "
        << Translate ("Enter the base URL of the archive") << " :<p>"
        << "<form method=\"get\" action=\"" << addarchive << "\">"
        << "<input type=text name=archiveurl size=50 maxlength=200>"
        << "<input type=hidden name=language value=\"" << language << "\">"
        << "<input type=hidden name=bgcolor value=\"" << bgcolor << "\">"
        << "<input type=hidden name=headercolor value=\"" << headercolor << "\">"
        << "<input type=hidden name=blockcolor value=\"" << blockcolor << "\">"
        << "<input type=hidden name=gotarchive value=\"gotarchive\">"
        << "<input type=hidden name=version value=\"oai20\">"
        << "<input type=submit name=submit value=\""
        << Translate ("Test the archive") << "\">"
        << "</form><hr>";

   cout << Translate ("If your archive supports OAI-PMH v1.1") << " ... "
        << Translate ("Enter the base URL of the archive") << " :<p>"
        << "<form method=\"get\" action=\"" << addarchive << "\">"
        << "<input type=text name=archiveurl size=50 maxlength=200>"
        << "<input type=hidden name=language value=\"" << language << "\">"
        << "<input type=hidden name=bgcolor value=\"" << bgcolor << "\">"
        << "<input type=hidden name=headercolor value=\"" << headercolor << "\">"
        << "<input type=hidden name=blockcolor value=\"" << blockcolor << "\">"
        << "<input type=hidden name=gotarchive value=\"gotarchive\">"
        << "<input type=hidden name=version value=\"oai11\">"
        << "<input type=submit name=submit value=\""
        << Translate ("Test the archive") << "\">"
        << "</form>";
}

// ----------------------------------------------------------------------
//  Function   : ScreenTwo
// ----------------------------------------------------------------------
//  Purpose    : test archive and generate screen for more data
//  Parameters : 1. archive url
//               2. language name
//               3. bgcolor
//               4. headercolor
//               5. blockcolor
//               6. protocol version
//  Result     : (none)
// ----------------------------------------------------------------------

void ScreenTwo ( char *archiveurl, char *language, char *bgcolor, 
                 char *headercolor, char *blockcolor, char *version )
{
   cout << "<pre>"; cout.flush ();
   
   char complycommand[2048];
   if (strcmp (version, "xoai10") == 0)
      sprintf (complycommand, "%s %s %s", complyxoai10path, archiveurl, language);
   else if (strcmp (version, "oai11") == 0)
      sprintf (complycommand, "%s %s %s", complyoai11path, archiveurl, language);
   else if (strcmp (version, "oai20") == 0)
      sprintf (complycommand, "%s %s %s", complyoai20path, archiveurl, language);
   int status = system (complycommand) >> 8;
   
   cout << "</pre>";
   
   if (status == 0)
   {
      char *archivename = GetProtocol (archiveurl);
   
      cout << "<hr><h3>" << Translate ("Congratulations !") << "</h3><p>"
           << Translate ("Your component satisfied all the tests we performed")
           << ". " << Translate ("You may now add it to the list of components on the front page if you wish.")
           << ":<p><form method=\"get\" action=\"" << addarchive << "\">"
           << "&nbsp;&nbsp;" << Translate ("Base URL of component") << " : " << archiveurl
           << "<p>&nbsp;&nbsp;" << Translate ("Type of component") << " : " << version
           << "<input type=hidden name=archiveurl value=\"" << archiveurl << "\">"
           << "<input type=hidden name=language value=\"" << language << "\">"
           << "<input type=hidden name=gotarchive value=\"gotarchive\">"
           << "<input type=hidden name=bgcolor value=\"" << bgcolor << "\">"
           << "<input type=hidden name=headercolor value=\"" << headercolor << "\">"
           << "<input type=hidden name=blockcolor value=\"" << blockcolor << "\">"
           << "<input type=hidden name=archivename value=\"" << archivename << "\">"
           << "<p>&nbsp;&nbsp;" << Translate ("Version/Name of Component") << " : " << archivename
           << "<p><input type=submit name=submit value=\"" << Translate ("Submit Component to List") << "\">"
           << "</form><hr>";
   }
}

// ----------------------------------------------------------------------
//  Function   : ScreenThree
// ----------------------------------------------------------------------
//  Purpose    : save information to predef file
//  Parameters : 1. archive oai url
//               2. archive name
//  Result     : (none)
// ----------------------------------------------------------------------

void ScreenThree ( char *archiveurl, char *archivename )
{
   fstream predef ("predef", ios::out|ios::app);
   
   predef << archivename << "\n";
   predef << archiveurl << "\n";
   
   cout << "<h4>" << Translate ("Thank You ! Your component has been added")
        << "</h4>";
}



int main ()
{
   CGIProcessor cgi;
   
   char *language = cgi.ParmList.Search ("language");
   char *archiveurl = cgi.ParmList.Search ("archiveurl");
   char *archivename = cgi.ParmList.Search ("archivename");
   char *archive = cgi.ParmList.Search ("archive");
   char *bgcolor = cgi.ParmList.Search ("bgcolor");
   char *headercolor = cgi.ParmList.Search ("headercolor");
   char *blockcolor = cgi.ParmList.Search ("blockcolor");   
   char *gotarchive = cgi.ParmList.Search ("gotarchive");
   char *version = cgi.ParmList.Search ("version");
   
   SetLanguage (language);
   if ((!language) || (strlen (language) == 0))
      language = "enus.lan";
   if ((!bgcolor) || (strlen (bgcolor) == 0))
      bgcolor = bgcolorpreset;
   if ((!headercolor) || (strlen (headercolor) == 0))
      headercolor = headercolorpreset;
   if ((!blockcolor) || (strlen (blockcolor) == 0))
      blockcolor = blockcolorpreset;

   OutputHeaders (reprotocolversion, bgcolor, headercolor);

   if ((archiveurl == NULL) || (strcmp (archiveurl, "") == 0) ||
       (gotarchive == NULL) || (strcmp (gotarchive, "") == 0))
      ScreenOne (language, bgcolor, headercolor, blockcolor);
   else if ((archivename == NULL) || (strcmp (archivename, "") == 0))
      ScreenTwo (archiveurl, language, bgcolor, headercolor, blockcolor, version);
   else
      ScreenThree (archiveurl, archivename);

   OutputFooters (language, bgcolor, headercolor, blockcolor);
   
   return 0;   
}
